@@ -1,5 +1,6 @@
 import logging
 import collections
+from functools import lru_cache
 
 import census
 from census.core import supported_years
@@ -12,10 +13,19 @@ from logging import NullHandler
 
 logging.getLogger(__name__).addHandler(NullHandler())
 
+class HashDict(dict):
+
+    def __hash__(self):
+        return hash(tuple(self.items()))
+
 
 class GeoClient(census.core.Client):
 
-    @supported_years(2019, 2018, 2017, 2016, 2015, 2014, 2013, 2012, 2011, 2010, 2000, 1990)
+    @lru_cache
+    def get(self, *args, **kwargs):
+        return super().get(*args, **kwargs)
+
+    @supported_years(2020, 2019, 2018, 2017, 2016, 2015, 2014, 2013, 2012, 2011, 2010, 2000)
     def geo_tract(self, fields, geojson_geometry, year=None, **kwargs):
         if year is None:
             year = self.default_year
@@ -30,8 +40,10 @@ class GeoClient(census.core.Client):
 
             tract_id = tract['properties']['TRACT']
             result = self.get(fields,
-                              {'for': 'tract:{}'.format(tract_id),
-                               'in':  within}, year, **kwargs)
+                              HashDict({'for': 'tract:{}'.format(tract_id),
+                                        'in': within}),
+                              year,
+                              **kwargs)
 
             if result:
                 result, = result
@@ -40,7 +52,7 @@ class GeoClient(census.core.Client):
 
             yield tract, result, intersection_proportion
 
-    @supported_years(2019, 2018, 2017, 2016, 2015, 2014, 2013, 2012, 2011, 2010, 2000)
+    @supported_years(2020, 2019, 2018, 2017, 2016, 2015, 2014, 2013, 2012, 2011, 2010)
     def geo_blockgroup(self, fields, geojson_geometry, year=None, **kwargs):
         if year is None:
             year = self.default_year
@@ -54,11 +66,14 @@ class GeoClient(census.core.Client):
                        'tract': block_group['properties']['TRACT']}
             within = 'state:{state} county:{county} tract:{tract}'.format(**context)
 
-            block_group_id = block_group['properties']['BLKGRP']
+            tract_blockgroups = self.get(fields,
+                                         HashDict({'for': 'block group:*',
+                                                   'in':  within}),
+                                         year,
+                                         **kwargs)
 
-            result = self.get(fields,
-                              {'for': 'block group:{}'.format(block_group_id),
-                               'in':  within}, year, **kwargs)
+            result = [result for result in tract_blockgroups
+                      if result['block group'] == block_group['properties']['BLKGRP']]
 
             if result:
                 result, = result
@@ -76,7 +91,14 @@ class GeoClient(census.core.Client):
                                            extra_query_args={'where': search_query,
                                                              'orderByFields': 'OID'})
 
-        place = next(iter(place_dumper))
+        try:
+            place = next(iter(place_dumper))
+        except TypeError as e:
+            if "'<' not supported between instances of 'NoneType' and 'NoneType'" in str(e):
+                raise ValueError(f'Could not find specified place "{place}" in state "{state}"')
+
+            raise e
+
         logging.info(place['properties']['NAME'])
         place_geojson = place['geometry']
 
@@ -162,20 +184,9 @@ class ACS5Client(census.core.ACS5Client, GeoClient):
         return self._state_place_area(self.geo_blockgroup, *args, **kwargs)
 
 
-class SF1Client(census.core.SF1Client, GeoClient):
-    @supported_years(2010)
-    def state_place_tract(self, *args, **kwargs):
-        return self._state_place_area(self.geo_tract, *args, **kwargs)
+class GeoBlockClient(GeoClient):
 
-    @supported_years(2010)
-    def state_place_blockgroup(self, *args, **kwargs):
-        return self._state_place_area(self.geo_blockgroup, *args, **kwargs)
-
-    @supported_years(2010)
-    def state_place_block(self, *args, **kwargs):
-        return self._state_place_area(self.geo_block, *args, **kwargs)
-
-    @supported_years(2010)
+    @supported_years(2020, 2010)
     def geo_block(self, fields, geojson_geometry, year):
         if year is None:
             year = self.default_year
@@ -189,17 +200,49 @@ class SF1Client(census.core.SF1Client, GeoClient):
                        'tract': block['properties']['TRACT']}
             within = 'state:{state} county:{county} tract:{tract}'.format(**context)
 
-            block_id = block['properties']['BLOCK']
-            result = self.get(fields,
-                              {'for': 'block:{}'.format(block_id),
-                               'in':  within}, year)
+            tract_blocks = self.get(fields,
+                                    HashDict({'for': 'block:*',
+                                              'in':  within}),
+                                    year)
+            
+            result = [result for result in tract_blocks
+                      if result['block'] == block['properties']['BLOCK']]
 
             if result:
                 result, = result
             else:
+                breakpoint()
                 result = {}
 
             yield block, result, intersection_proportion
+    
+
+class SF1Client(census.core.SF1Client, GeoBlockClient):
+    @supported_years(2010)
+    def state_place_tract(self, *args, **kwargs):
+        return self._state_place_area(self.geo_tract, *args, **kwargs)
+
+    @supported_years(2010)
+    def state_place_blockgroup(self, *args, **kwargs):
+        return self._state_place_area(self.geo_blockgroup, *args, **kwargs)
+
+    @supported_years(2010)
+    def state_place_block(self, *args, **kwargs):
+        return self._state_place_area(self.geo_block, *args, **kwargs)
+
+            
+class PLClient(census.core.PLClient, GeoBlockClient):
+    @supported_years(2020, 2010)
+    def state_place_tract(self, *args, **kwargs):
+        return self._state_place_area(self.geo_tract, *args, **kwargs)
+
+    @supported_years(2020, 2010)
+    def state_place_blockgroup(self, *args, **kwargs):
+        return self._state_place_area(self.geo_blockgroup, *args, **kwargs)
+
+    @supported_years(2020, 2010)
+    def state_place_block(self, *args, **kwargs):
+        return self._state_place_area(self.geo_block, *args, **kwargs)
 
 
 class Census(census.Census):
@@ -207,3 +250,4 @@ class Census(census.Census):
         super(Census, self).__init__(key, year, session)
         self.acs5 = ACS5Client(key, year, session)
         self.sf1 = SF1Client(key, year, session)
+        self.pl = PLClient(key, year, session)
